from starlette.requests import Request
from fasteve.methods import delete_item, get_item, patch_item, put_item
from fastapi import HTTPException
from fastapi import Path
from typing import Callable, Optional, Union
from fasteve.resource import Resource
from fasteve.core.utils import ObjectID
from pymongo.errors import DuplicateKeyError, BulkWriteError
from fasteve.io.mongo.utils import render_pymongo_error


async def process_item_request(
    request: Request, item_id: Union[ObjectID, str]
) -> Optional[dict]:
    methods = {
        "GET": get_item,
        "DELETE": delete_item,
        "PUT": put_item,
        "PATCH": patch_item,
    }
    if request.method not in methods:
        raise HTTPException(405)
    try:
        res = await methods[request.method](request, item_id)  # type: ignore
    except DuplicateKeyError as e:
        msg = render_pymongo_error(e.details)
        raise HTTPException(422, msg)
    except BulkWriteError as e:
        msg = render_pymongo_error(e.details["writeErrors"][0])
        raise HTTPException(422, msg)
    except Exception as e:
        raise e
    return res


def item_endpoint_factory(resource: Resource, method: str) -> Callable:
    """Dynamically create item endpoint"""

    if method in ("GET", "HEAD", "DELETE"):
        # no schema validation on GET request
        # TODO: DELETE needs if-match header?
        if resource.alt_id:

            async def item_endpoint_with_alt_id(
                request: Request,
                item_id: Union[ObjectID, str] = Path(
                    ..., alias=f"{resource.item_name}_id"
                ),
            ) -> Optional[dict]:
                return await process_item_request(request, item_id)

            return item_endpoint_with_alt_id
        else:

            async def item_endpoint(
                request: Request,
                item_id: ObjectID = Path(..., alias=f"{resource.item_name}_id"),
            ) -> Optional[dict]:
                response = await process_item_request(request, item_id)
                await request.app.events.run(
                    "after_fetch_item", resource.name, response
                )
                return response

            return item_endpoint
    elif method in ["PUT", "PATCH"]:
        schema = resource.schema
        if method == "PATCH":
            schema = type(
                schema.__name__,
                (schema,),
                {},
            )
            fields = schema.__fields__
            for field_name in fields:
                fields[field_name].required = False
            # schema.__fields__ = fields
        if resource.alt_id:

            async def mod_item_endpoint_with_alt_id(
                request: Request,
                schema: schema,  # type: ignore
                item_id: Union[ObjectID, str] = Path(
                    ..., alias=f"{resource.item_name}_id"
                ),
            ) -> Optional[dict]:
                setattr(request, "payload", schema.dict(exclude_unset=True))  # type: ignore
                return await process_item_request(request, item_id)

            return mod_item_endpoint_with_alt_id
        else:

            async def mod_item_endpoint(
                request: Request,
                schema: schema,  # type: ignore
                item_id: ObjectID = Path(..., alias=f"{resource.item_name}_id"),
            ) -> Optional[dict]:
                setattr(request, "payload", schema.dict(exclude_unset=True))  # type: ignore
                return await process_item_request(request, item_id)

            return mod_item_endpoint
    else:
        raise Exception(f'"{method}" is an invalid HTTP method')

    return item_endpoint



import hashlib
from json import dumps
from typing import Any, List
from .resource import Resource
from pydantic import Field


def document_etag(value: dict, ignore_fields: List[str] = None) -> str:
    """Computes and returns a valid ETag for the input value."""
    h = hashlib.sha1()
    h.update(dumps(value, sort_keys=True).encode("utf-8"))
    return h.hexdigest()


def Unique(old_type: Any) -> Any:
    Unique = type("Unique", (old_type,), {"__name__": "Unique"})
    return Unique


def DataRelation(resource: Resource, optional: bool = True) -> Any:
    # the optional flag is dumb. figure out a better way to do it.
    # basically i just want to use Option[] on the type
    #
    if optional:
        Relation = Field(None, data_relation=resource)
    else:
        Relation = Field(..., data_relation=resource)
    return Relation



from fastapi import FastAPI, APIRouter
from .middleware.resource import ResourceMiddleware
from .middleware.cors import CORSMiddleware
from .endpoints import (
    collections_endpoint_factory,
    home_endpoint,
    item_endpoint_factory,
    subresource_endpoint_factory,
)
from .core import config
from .events import Events
from .schema import BaseResponseSchema, BaseSchema, ItemBaseResponseSchema
from typing import List, Type, Optional, Union, Callable
from types import ModuleType
from .resource import Resource
from .io.mongo import Mongo, MongoClient
from .io.base import DataLayer
from .core.utils import log, ObjectID, is_new_type, repeat_every as repeat
from datetime import datetime
from pydantic import Field, create_model, BaseModel
import asyncio
import logging


class Fasteve(FastAPI):
    def __init__(
        self,
        resources: List[Resource] = [],
        data: Type[DataLayer] = Mongo,
        cors_origins: List[str] = [],
    ) -> None:

        super().__init__()  # Initialise FastAPI super class

        # set defaults
        self.cors_origins = cors_origins

        # validate user settings
        self._validate_resources(resources)
        self.resources = resources

        self._validate_config(config)
        self.config = config

        self._register_resource_middleware()
        self._register_CORS_middleware()
        self._register_home_endpoint()

        # connect to db
        MongoClient.connect()

        self.data = data(app=self)  # eve pattern

        for resource in self.resources:
            self.create_mongo_index(resource)

        for resource in self.resources:
            self.register_resource(resource)

        self.events = Events(resources)
        setattr(self.router, "add_event_handler", self.add_event_handler)
        self.add_event_handler(
            "shutdown", MongoClient.close
        )  # this can't be in the application layer i.e. needs to come from data layer

    def add_event_handler(self, event_type: str, func: Callable) -> None:
        if event_type == "startup":
            self.router.on_startup.append(func)
        elif event_type == "shutdown":
            self.router.on_shutdown.append(func)
        else:
            # Event Hooks
            try:
                getattr(self.events, event_type).append(func)
            except Exception as e:
                raise e

    @log
    def register_resource(self, resource: Resource) -> None:
        print(f"Registering Resource: {resource.name}")
        # process resource_settings
        # add name to api
        router = APIRouter()

        # check models for data relations
        resource.schema = self._embed_data_relation(resource.schema)
        resource.schema.__config__.extra = (  # type: ignore
            "forbid"  # type: ignore # TODO: this should be on the InSchema
        )
        resource.response_model = self._embed_data_relation(
            resource.response_model, response=True
        )

        response_model = self._prepare_response_model(
            resource.response_model, resource.name
        )

        Response = create_model(
            f"ResponseSchema_{resource.name}",
            data=(List[response_model], Field(..., alias="_data")),  # type: ignore
            __base__=BaseResponseSchema,
        )

        PostResponse = create_model(
            f"PostResponseSchema_{resource.name}",
            data=(List[response_model], Field(..., alias="_data")),  # type: ignore
            __base__=BaseSchema,  # No meta or links
        )

        for method in resource.resource_methods:
            if method == "POST":
                router.add_api_route(
                    f"/{resource.name}",
                    endpoint=collections_endpoint_factory(resource, method),
                    response_model=PostResponse,
                    response_model_exclude_unset=True,
                    methods=[method],
                    status_code=201,
                )
            elif method == "DELETE":
                router.add_api_route(
                    f"/{resource.name}",
                    endpoint=collections_endpoint_factory(resource, method),
                    methods=[method],
                    status_code=204,
                )
            else:
                router.add_api_route(
                    f"/{resource.name}",
                    endpoint=collections_endpoint_factory(resource, method),
                    response_model=Response,
                    response_model_exclude_unset=True,
                    methods=[method],
                )

        ItemResponse = create_model(
            f"ItemResponseSchema_{resource.name}",
            data=(List[response_model], Field(..., alias="_data")),  # type: ignore
            __base__=ItemBaseResponseSchema,
        )

        for method in resource.item_methods:
            if method in ["PUT", "DELETE", "PATCH"]:
                router.add_api_route(
                    f"/{resource.name}/{{{str(resource.item_name) + '_id'}}}",
                    endpoint=item_endpoint_factory(resource, method),
                    methods=[method],
                    status_code=204,
                )
            else:
                router.add_api_route(
                    f"/{resource.name}/{{{str(resource.item_name) + '_id'}}}",
                    endpoint=item_endpoint_factory(resource, method),
                    response_model=ItemResponse,
                    response_model_exclude_unset=True,
                    methods=[method],
                )

        for sub_resource in resource.sub_resources:
            Response = create_model(
                f"ResponseSchema_{resource.name}_sub_resource_{sub_resource.name}",
                data=(List[sub_resource.resource.response_model], Field(..., alias="_data")),  # type: ignore
                __base__=BaseResponseSchema,
            )

            print(f"Registering Sub Resource: {sub_resource.name}")
            router.add_api_route(
                f"/{resource.name}/{{{str(resource.item_name) + '_id'}}}/{sub_resource.name}",
                endpoint=subresource_endpoint_factory(resource, "GET", sub_resource),
                response_model=Response,
                response_model_exclude_unset=True,
                methods=["GET"],
            )

        # TODO: api versioning
        self.include_router(
            router,
            tags=[str(resource.name)],
        )

    async def _create_mongo_index_internal(
        self, resource: Resource, field_name: str
    ) -> None:
        collection = await self.data.get_collection(resource)
        # TODO: index should be removed at some point...
        # or at least check to see if there is one already
        res = await collection.create_index(field_name, unique=True)  # type: ignore # TODO: move to data layer
        print(f"Created unique index for {field_name} in {resource.name}")

    def create_mongo_index(self, resource: Resource) -> None:
        fields = resource.schema.__fields__
        for name in fields:
            field = fields[name]
            type_ = field.type_
            if not is_new_type(type_):
                continue
            if type_.__name__ == "Unique":
                asyncio.create_task(self._create_mongo_index_internal(resource, name))

    def repeat_every(
        self,
        *,
        seconds: float,
        wait_first: bool = False,
        logger: Optional[logging.Logger] = None,
        raise_exceptions: bool = False,
        max_repetitions: Optional[int] = None,
    ) -> Callable:
        """https://stackoverflow.com/a/52518284/5209891"""
        dec2 = repeat(
            seconds=seconds,
            wait_first=wait_first,
            logger=logger,
            raise_exceptions=raise_exceptions,
            max_repetitions=max_repetitions,
        )
        dec1 = self.on_event("startup")

        def merged_decorator(func: Callable) -> Callable:
            return dec1(dec2(func))

        return merged_decorator

    def _embed_data_relation(
        self, model: Type[BaseModel], response: bool = False
    ) -> Type[BaseModel]:

        fields = model.__fields__.keys()
        for name in fields:
            field = model.__fields__[name]

            if not "data_relation" in field.field_info.extra:
                continue

            data_relation = field.field_info.extra["data_relation"]
            outer_type_ = field.outer_type_
            many = False

            if outer_type_ not in (ObjectID, List[ObjectID]):
                raise ValueError(
                    f"Data relation ({model.__name__}: {name}) must be must be type ObjectID or List[ObjectID]"
                )

            if response:
                response_model = self._prepare_response_model(
                    data_relation.response_model, data_relation.name + "_embedded"
                )  # add ids, create, updated, ect
                if outer_type_ == ObjectID:
                    type_ = Union[Type[outer_type_], Type[response_model]]  # type: ignore # meta typing
                else:
                    type_ = Union[Type[outer_type_], List[Type[response_model]]]  # type: ignore # meta typing
                    many = True
            else:
                type_ = outer_type_  # type: ignore # meta typing

            if field.required:
                relation_field = {
                    name: (type_, Field(..., data_relation=data_relation, many=many))
                }
            else:
                relation_field = {
                    name: (type_, Field(None, data_relation=data_relation, many=many))
                }

            # relation_field = {name:(type_,field)}
            model = create_model(model.__name__, **relation_field, __base__=model)  # type: ignore # TODO: define multi-type
        return model

    def _prepare_response_model(
        self, response_model: Type[BaseModel], name: str
    ) -> Type[BaseModel]:
        return create_model(
            f"out_schema_{name}",
            id=(ObjectID, Field(..., alias="_id")),
            updated=(datetime, Field(..., alias="_updated")),
            created=(datetime, Field(..., alias="_created")),
            __base__=response_model,
        )

    def _register_home_endpoint(self) -> None:
        self.add_api_route(f"/", home_endpoint)

    def _register_resource_middleware(self) -> None:
        """Pass resources to every request"""
        # app is already passed to every request!
        # therfore I can use request.app.resources
        self.add_middleware(ResourceMiddleware, resources=self.resources)

    def _register_CORS_middleware(self) -> None:
        """Set global Cross-Origin Resource Sharing"""
        if self.cors_origins:
            # app level cors
            origins_raw = self.cors_origins
        elif config.CORS_ORIGINS:
            # global cors
            origins_raw = config.CORS_ORIGINS.split(",")
        # CORS
        origins = []
        # Set all CORS enabled origins
        if config.CORS_ORIGINS:
            for origin in origins_raw:
                use_origin = origin.strip()
                origins.append(use_origin)
            self.add_middleware(
                CORSMiddleware,
                allow_origins=origins,
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

    def _validate_config(self, config: ModuleType) -> None:
        # raise errors if config is invalid
        pass

    def _validate_resources(self, resources: List[Resource]) -> None:
        # raise errors if resource is invalid
        for resource in resources:
            # check methods i.e. post only on resource
            # check alt_id is unique
            pass



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional, Type
from pydantic import BaseModel


@dataclass
class Resource:
    schema: Type[BaseModel]  # in the db
    name: str = ""
    item_name: str = ""
    resource_methods: List[str] = field(default_factory=lambda: ["GET"])
    item_methods: List[str] = field(default_factory=lambda: ["GET"])
    response_model: Type[BaseModel] = None  # type: ignore # schema used as default
    response_model_include: set = field(default_factory=lambda: set())  # TODO
    response_model_exclude: set = field(default_factory=lambda: set())  # TODO
    alt_id: Optional[str] = None
    sub_resources: List[SubResource] = field(default_factory=lambda: list())

    allowed_filters: bool = True
    projection: bool = True
    sorting: bool = True
    embedding: bool = True
    datasource: Optional[dict] = None
    bulk_inserts: bool = True

    def __post_init__(self) -> None:
        if not self.name:
            self.name = self.schema.__name__.lower()
        if not self.item_name:
            if self.name.endswith("s"):
                self.item_name = self.name[:-1]
            else:
                self.item_name = self.name
        if not self.response_model:
            self.response_model = self.schema


@dataclass
class SubResource:
    # sub resource should set the id on the parent automajically
    resource: Resource
    id_field: str
    name: Optional[str] = None

    def __post_init__(self) -> None:
        if not self.name:
            self.name = self.resource.schema.name  # type: ignore # set name to resource schema name by default



from starlette.requests import Request
from fasteve.methods import delete, get, post
from fastapi import HTTPException
from typing import Callable, List, Union
from fasteve.resource import Resource
from fasteve.core.utils import log
from pymongo.errors import DuplicateKeyError, BulkWriteError
from fasteve.io.mongo.utils import render_pymongo_error


@log
async def process_collections_request(request: Request) -> dict:
    methods: dict = {"GET": get, "POST": post, "DELETE": delete}
    if request.method not in methods:
        raise HTTPException(405)
    try:
        res = await methods[request.method](request)
        return res
    except DuplicateKeyError as e:
        msg = render_pymongo_error(e.details)
        raise HTTPException(422, msg)
    except BulkWriteError as e:
        msg = render_pymongo_error(e.details["writeErrors"][0])
        raise HTTPException(422, msg)
    except Exception as e:
        raise e


def collections_endpoint_factory(resource: Resource, method: str) -> Callable:
    """Dynamically create collection endpoints"""
    if method in ("GET", "HEAD"):
        # no schema validation on GET request
        # query prams
        async def get_endpoint(
            request: Request, max_results: int = 25, page: int = 1, embedded: str = "{}"
        ) -> dict:
            response = await process_collections_request(request)
            await request.app.events.run(
                "after_fetch_resource", resource.name, response
            )
            return response

        return get_endpoint

    elif method == "POST":
        if resource.bulk_inserts:
            schema = Union[List[resource.schema], resource.schema]  # type: ignore
        else:
            schema = resource.schema  # type: ignore

        async def post_endpoint(request: Request, schema: schema) -> dict:
            payload = (
                [schema.dict() for schema in schema]  # type: ignore
                if type(schema) == list
                else [schema.dict()]  # type: ignore
            )

            setattr(request, "payload", payload)
            await request.app.events.run("before_insert_items", resource.name, payload)
            response = await process_collections_request(request)
            await request.app.events.run("after_insert_items", resource.name, response)
            return response

        return post_endpoint

    elif method == "DELETE":
        # no schema validation on DELETE HEAD request
        async def delete_endpoint(
            request: Request,
        ) -> dict:
            return await process_collections_request(request)

        return delete_endpoint

    else:
        raise Exception(f'"{method}" is an invalid HTTP method')



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response
from starlette.types import ASGIApp
from typing import Callable, List
from starlette.requests import Request
from ..resource import Resource


class ResourceMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, resources: List[Resource]) -> None:
        self.app = app
        self.dispatch_func = self.dispatch
        self.resources = resources

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        if request.scope["path"] != "/":
            # Use pattern matching to find resource
            route = request.scope["path"][1:].split("/")[0]
            resource = next(
                (
                    resource
                    for resource in self.resources
                    if resource.name is not None
                    and resource.name.lower() == route.lower()
                ),
                None,
            )
        else:
            resource = None
        request.state.resource = resource
        if resource:
            await request.app.events.run(
                f"before_{request.method}", resource.name, request
            )
        response = await call_next(request)
        if resource:
            await request.app.events.run(
                f"after_{request.method}", resource.name, request, response
            )
        return response



from starlette.config import Config

# Config will be read from environment variables and/or ".env" files.
# Should envirmonet variables have priority?
config = Config(".env")

INFO = config("INFO", cast=str, default="_info")
API_VERSION = config("API_VERSION", cast=str, default="0.1")
HATEOAS = config("HATEOAS", cast=bool, default=True)
LINKS = config("LINKS", cast=str, default="_links")
META = config("META", cast=str, default="_meta")
DATA = config("DATA", cast=str, default="_data")
PAGINATION = config("PAGINATION", cast=bool, default=True)
MONGODB_URI = config("MONGODB_URI", cast=str, default="mongodb://localhost:27017")
MONGODB_DATABASE = config("MONGODB_DATABASE", cast=str, default="fasteve_db")
CORS_ORIGINS = config("CORS_ORIGINS", cast=str, default=None)
CONNECTION_TIMEOUT = config("CONNECTION_TIMEOUT", cast=int, default=10000)

# default query parameters
QUERY_WHERE = config("QUERY_WHERE", cast=str, default="where")
QUERY_PROJECTION = config("QUERY_PROJECTION", cast=str, default="projection")
QUERY_SORT = config("QUERY_SORT", cast=str, default="sort")
QUERY_PAGE = config("QUERY_PAGE", cast=str, default="page")
QUERY_MAX_RESULTS = config("QUERY_MAX_RESULTS", cast=str, default="max_results")
QUERY_EMBEDDED = config("QUERY_EMBEDDED", cast=str, default="embedded")
QUERY_AGGREGATION = config("QUERY_AGGREGATION", cast=str, default="aggregate")



# -*- coding: utf-8 -*-

"""
    eve.io.base
    ~~~~~~~~~~~

    Standard interface implemented by Eve data layers.

    :copyright: (c) 2017 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from typing import Optional, List, Tuple
from fasteve.resource import Resource


class ConnectionException(Exception):
    """Raised when DataLayer subclasses cannot find/activate to their
    database connection.

    :param driver_exception: the original exception raised by the source db
                             driver
    """

    def __init__(self, driver_exception: Optional[str] = None):
        self.driver_exception = driver_exception

    def __str__(self) -> str:
        msg = (
            "Error initializing the driver. Make sure the database server"
            "is running. "
        )
        if self.driver_exception:
            msg += "Driver exception: %s" % repr(self.driver_exception)
        return msg


class Client:
    """Database client where you connect to the database"""

    async def get_database(self) -> None:
        raise NotImplementedError

    def connect(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError


class DataLayer:
    """Base data layer class. Defines the interface that actual data-access
    classes, being subclasses, must implement. Implemented as a Flask
    extension.

    Admittedly, this interface is a Mongo rip-off. See the io.mongo
    package for an implementation example.
    """

    class OriginalChangedError(Exception):
        pass

    # if custom serialize functions are needed, add them to the 'serializers'
    # dictionary, eg:
    # serializers = {'objectid': ObjectId, 'datetime': serialize_date}
    serializers: dict = {}

    def __init__(self, app) -> None:  # type: ignore
        """Implements extension pattern."""
        if app is not None:
            self.app = app
            self.init_app()
        else:
            self.app = None

    def init_app(self) -> None:
        """This is where you want to initialize the db driver so it will be
        alive through the whole instance lifespan.
        """
        raise NotImplementedError

    async def get_collection(self, resource: Resource) -> None:
        raise NotImplementedError

    def connect(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        raise NotImplementedError

    async def find(self, resource: Resource, args: dict) -> Tuple[List[dict], int]:
        """Retrieves a set of documents (rows), matching the current request.
        Consumed when a request hits a collection/document endpoint
        (`/people/`).

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve both
                         the db collection/table and base query (filter), if
                         any.
        :param req: an instance of ``eve.utils.ParsedRequest``. This contains
                    all the constraints that must be fulfilled in order to
                    satisfy the original request (where and sort parts, paging,
                    etc). Be warned that `where` and `sort` expressions will
                    need proper parsing, according to the syntax that you want
                    to support with your driver. For example ``eve.io.Mongo``
                    supports both Python and Mongo-like query syntaxes.
        :param sub_resource_lookup: sub-resource lookup from the endpoint url.
        :param perform_count: wether a document count should be performed and
                              returned to the client.
        """
        raise NotImplementedError

    async def aggregate(
        self,
        resource: Resource,
        pipline: List[dict] = [],
        skip: int = 0,
        limit: int = 0,
    ) -> Tuple[List[dict], int]:
        """Perform an aggregation on the resource datasource and returns
        the result. Only implent this if the underlying db engine supports
        aggregation operations.

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve
                         the db collection/table consumed by the resource.
        :param pipeline: aggregation pipeline to be executed.
        :param options: aggregation options to be considered.
        """
        raise NotImplementedError

    async def find_one(self, resource: Resource, lookup: dict) -> dict:
        """Retrieves a single document/record. Consumed when a request hits an
        item endpoint (`/people/id/`).

        :param resource: resource being accessed. You should then use the
                         ``datasource`` helper function to retrieve both the
                         db collection/table and base query (filter), if any.
        :param req: an instance of ``eve.utils.ParsedRequest``. This contains
                    all the constraints that must be fulfilled in order to
                    satisfy the original request (where and sort parts, paging,
                    etc). As we are going to only look for one document here,
                    the only req attribute that you want to process here is
                    ``req.projection``.
        :param check_auth_value: a boolean flag indicating if the find
                                 operation should consider user-restricted
                                 resource access. Defaults to ``True``.
        :param force_auth_field_projection: a boolean flag indicating if the
                                            find operation should always
                                            include the user-restricted
                                            resource access field (if
                                            configured). Defaults to ``False``.

        :param **lookup: the lookup fields. This will most likely be a record
                         id or, if alternate lookup is supported by the API,
                         the corresponding query.
        """
        raise NotImplementedError

    def find_one_raw(self, resource: Resource, **lookup: dict) -> None:
        """Retrieves a single, raw document. No projections or datasource
        filters are being applied here. Just looking up the document using the
        same lookup.

        :param resource: resource name.
        :param ** lookup: lookup query.
        """
        raise NotImplementedError

    def find_list_of_ids(self, resource: Resource, ids: List) -> None:
        """Retrieves a list of documents based on a list of primary keys
        The primary key is the field defined in `ID_FIELD`.
        This is a separate function to allow us to use per-database
        optimizations for this type of query.

        :param resource: resource name.
        :param ids: a list of ids corresponding to the documents
        to retrieve
        :param client_projection: a specific projection to use
        :return: a list of documents matching the ids in `ids` from the
        collection specified in `resource`
        """
        raise NotImplementedError

    def insert(self, resource: Resource, payload: dict) -> dict:
        """Inserts a document into a resource collection/table.

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve both
                         the actual datasource name.
        :param doc_or_docs: json document or list of json documents to be added
                            to the database.
        """
        raise NotImplementedError

    async def insert_many(self, resource: Resource, payload: List[dict]) -> List[dict]:
        """Inserts a document into a resource collection/table.

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve both
                         the actual datasource name.
        :param doc_or_docs: json document or list of json documents to be added
                            to the database.
        """
        raise NotImplementedError

    def update(
        self, resource: Resource, id_: str, updates: dict, original: dict
    ) -> None:
        """Updates a collection/table document/row.
        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve
                         the actual datasource name.
        :param id_: the unique id of the document.
        :param updates: json updates to be performed on the database document
                        (or row).
        :param original: definition of the json document that should be
        updated.
        :raise OriginalChangedError: raised if the database layer notices a
        change from the supplied `original` parameter.
        """
        raise NotImplementedError

    def replace(
        self, resource: Resource, id_: str, document: dict, original: dict
    ) -> None:
        """Replaces a collection/table document/row.
        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve
                         the actual datasource name.
        :param id_: the unique id of the document.
        :param document: the new json document
        :param original: definition of the json document that should be
        updated.
        :raise OriginalChangedError: raised if the database layer notices a
        change from the supplied `original` parameter.
        """
        raise NotImplementedError

    async def remove(self, resource: Resource) -> None:
        """Removes a document/row or an entire set of documents/rows from a
        database collection/table.

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve
                         the actual datasource name.
        :param lookup: a dict with the query that documents must match in order
                       to qualify for deletion. For single document deletes,
                       this is usually the unique id of the document to be
                       removed.
        """
        raise NotImplementedError

    def combine_queries(self, query_a: dict, query_b: dict) -> None:
        """Takes two db queries and applies db-specific syntax to produce
        the intersection.
        """
        raise NotImplementedError

    def get_value_from_query(self, query: dict, field_name: str) -> None:
        """Parses the given potentially-complex query and returns the value
        being assigned to the field given in `field_name`.

        This mainly exists to deal with more complicated compound queries
        """
        raise NotImplementedError

    def query_contains_field(self, query: dict, field_name: str) -> None:
        """For the specified field name, does the query contain it?
        Used know whether we need to parse a compound query.
        """
        raise NotImplementedError

    def is_empty(self, resource: Resource) -> None:
        """Returns True if the collection is empty; False otherwise. While
        a user could rely on self.find() method to achieve the same result,
        this method can probably take advantage of specific datastore features
        to provide better performance.

        Don't forget, a 'resource' could have a pre-defined filter. If that is
        the case, it will have to be taken into consideration when performing
        the is_empty() check (see eve.io.mongo.mongo.py implementation).

        :param resource: resource being accessed. You should then use
                         the ``datasource`` helper function to retrieve
                         the actual datasource name.
        """
        raise NotImplementedError



from starlette.requests import Request
from datetime import datetime
from fasteve.methods.post import post
from fasteve.methods.common import get_document
from typing import Union
from fasteve.core.utils import ObjectID
from fastapi import Response


async def put_item(request: Request, item_id: Union[ObjectID, str]) -> Response:
    orginal_document = await get_document(request, item_id)
    payload = getattr(request, "payload")

    if not orginal_document:
        # insert
        if ObjectID.is_valid(item_id):
            payload["_id"] = item_id
        # if it is not valid must be an alt_id
        setattr(request, "payload", [payload])  # post expects a list
        await post(request)
        return Response(status_code=204)

    # replace
    now = datetime.now()
    payload["_created"] = orginal_document["_created"]
    payload["_updated"] = now

    try:
        document = await request.app.data.replace_item(
            request.state.resource, orginal_document["_id"], payload
        )
    except Exception as e:
        raise e
    return Response(status_code=204)



from starlette.requests import Request
from fasteve.core.utils import log, ObjectID
from fasteve.core import config
from math import ceil
from fastapi import HTTPException
from typing import Union
import json


@log
async def get(request: Request) -> dict:
    resource = request.state.resource
    query_params = dict(request.query_params)
    path_params = dict(request.path_params)
    limit = int(query_params["max_results"]) if "max_results" in query_params else 25
    page = int(query_params["page"]) if "page" in query_params else 1
    skip = (page - 1) * limit if page > 1 else 0

    query = {}
    pipeline = []

    item_id = (
        path_params[f"{resource.item_name}_id"]
        if f"{resource.item_name}_id" in path_params
        else None
    )
    if item_id:
        # subresource
        subresource_path_name = request.url.path.split("/")[-1]
        sub_resource = next(
            (
                resource
                for resource in resource.sub_resources
                if resource.name == subresource_path_name
            ),
            None,
        )
        if not sub_resource:
            raise HTTPException(404)
        try:
            query[sub_resource.id_field] = ObjectID.validate(item_id)
        except ValueError:
            lookup = {}
            lookup["from"] = resource.name
            lookup["localField"] = sub_resource.id_field
            lookup["foreignField"] = "_id"
            lookup["as"] = sub_resource.id_field
            pipeline.append({"$lookup": lookup})
            pipeline.append({"$unwind": "$" + sub_resource.id_field})
            pipeline.append(
                {"$match": {f"{sub_resource.id_field}.{resource.alt_id}": item_id}}
            )
            pipeline.append(
                {"$addFields": {sub_resource.id_field: f"${sub_resource.id_field}._id"}}
            )

        resource = sub_resource.resource
    try:
        if "embedded" in query_params and query_params["embedded"]:
            embedded = json.loads(query_params["embedded"])
        else:
            embedded = {}
    except:
        detail = [
            {
                "loc": ["query", "embedded"],
                "msg": "value is not a valid dict",
                "type": "type_error.dict",
            }
        ]
        raise HTTPException(422, detail)

    for field_name in embedded:
        if not embedded[field_name]:
            continue
        try:
            field = resource.response_model.__fields__[field_name]
        except:
            detail = [
                {
                    "loc": ["query", "embedded"],
                    "msg": f"field '{field_name}' is not valid",
                    "type": "value_error.not_valid",
                }
            ]
            raise HTTPException(422, detail)
        try:
            field.field_info.extra["data_relation"]
        except:
            detail = [
                {
                    "loc": ["query", "embedded"],
                    "msg": f"field '{field_name}' is not a embedable",
                    "type": "value_error.not_embedable",
                }
            ]
            raise HTTPException(422, detail)
        lookup = {}
        lookup["from"] = field_name
        lookup["localField"] = field_name
        lookup["foreignField"] = "_id"
        lookup["as"] = field_name
        stage = {"$lookup": lookup}
        pipeline.append(stage)  # assumes valid
        if not field.field_info.extra["many"]:
            pipeline.append({"$unwind": "$" + field_name})
    if pipeline:
        documents, count = await request.app.data.aggregate(
            resource, pipline=pipeline, skip=skip, limit=limit
        )
    else:
        try:
            documents, count = await request.app.data.find(
                resource, query=query, skip=skip, limit=limit
            )
        except Exception as e:
            raise e
    response = {}

    response[config.DATA] = documents

    if config.PAGINATION:
        response[config.META] = {
            "page": page,
            "max_results": limit,
            "total": count,
        }  # _meta_links(req, count)

    if config.HATEOAS:
        max_results = "&max_result=" + str(limit) if limit != 25 else ""
        response[config.LINKS] = {
            "self": {"href": request["path"], "title": resource.name},
            "parent": {"href": "/", "title": "home"},
        }  # _pagination_links(resource, req, count)
        if count > limit:
            response[config.LINKS]["next"] = {
                "href": f"{request['path']}?page={page + 1}{max_results}",
                "title": "next page",
            }
            response[config.LINKS]["last"] = {
                "href": f"{request['path']}?page={ceil(count / limit)}{max_results}",
                "title": "last page",
            }
    return response


async def get_item(request: Request, item_id: Union[ObjectID, str]) -> dict:
    try:
        query = {"_id": ObjectID.validate(item_id)}
    except ValueError:
        query = {request.state.resource.alt_id: item_id}
    try:
        item = await request.app.data.find_one(request.state.resource, query)
    except Exception as e:
        raise e
    if not item:
        raise HTTPException(404)
    response = {}

    response[config.DATA] = [item]
    return response



from pydantic import BaseModel, Field
from typing import Optional
from bson import ObjectId

json_encoders = {ObjectId: lambda x: str(x)}


class BaseSchema(BaseModel):
    class Config:
        json_encoders = json_encoders


class MetaModel(BaseModel):
    page: Optional[int]
    max_results: Optional[int]
    total: Optional[int]


class LinkModel(BaseModel):
    href: str
    title: str


class LinksModel(BaseModel):
    self: Optional[LinkModel]
    parent: Optional[LinkModel]
    next: Optional[LinkModel]
    last: Optional[LinkModel]


class ItemLinksModel(BaseModel):
    self: Optional[LinkModel]
    parent: Optional[LinkModel]


class BaseResponseSchema(BaseSchema):
    meta: Optional[MetaModel] = Field(
        "MetaModel",
        alias="_meta",
    )
    links: Optional[LinksModel] = Field("LinksModel", alias="_links")


class ItemBaseResponseSchema(BaseSchema):
    links: Optional[ItemLinksModel] = Field("ItemLinksModel", alias="_links")



from starlette.requests import Request
from datetime import datetime
from ..core import config


async def post(request: Request) -> dict:

    payload = getattr(request, "payload")

    now = datetime.now()

    for item in payload:
        item["_created"] = now
        item["_updated"] = now

    if len(payload) > 1:
        try:
            # TODO: bulk insert
            documents = await request.app.data.insert_many(
                request.state.resource, payload
            )
        except Exception as e:
            raise e
    else:
        try:
            payload = payload[0]
            document = await request.app.data.insert(request.state.resource, payload)
            documents = [document]
        except Exception as e:
            raise e

    response = {}

    response[config.DATA] = documents

    return response



from inspect import iscoroutinefunction


class Events:
    def __init__(self, resources: list) -> None:
        self.METHODS = ["GET", "HEAD", "POST", "DELETE", "PUT", "PATCH"]
        self.resources = resources

        # request events
        for method in self.METHODS:
            setattr(self, f"before_{method}", [])
            setattr(self, f"after_{method}", [])

        for resource in self.resources:
            for method in self.METHODS:
                setattr(self, f"before_{method}_{resource.name}", [])
                setattr(self, f"after_{method}_{resource.name}", [])

        # db events
        self.add("after", "fetch", levels=["resource", "item"])
        self.add("before", "insert", ["items"])  # resource?
        self.add("after", "insert", ["items"])  # resource?
        self.add("before", "replace", ["item"])
        self.add("after", "replace", ["item"])
        self.add("before", "update", ["item"])
        self.add("after", "update", ["item"])
        self.add("before", "delete", levels=["resource", "item"])
        self.add("after", "delete", levels=["resource", "item"])

    def add(self, timing: str, action: str, levels: list) -> None:
        assert timing in ("before", "after")

        for level in levels:
            setattr(self, f"{timing}_{action}_{level}", [])

        for resource in self.resources:
            for level in levels:
                setattr(self, f"{timing}_{action}_{level}_{resource.name}", [])

    async def run_callbacks(self, callbacks: list, *args: str) -> None:
        for func in callbacks:
            if iscoroutinefunction(func):
                await func(*args)
            else:
                func(*args)

    async def run(self, event: str, resource_name: str, *args: str) -> None:
        callbacks = getattr(self, event)
        await self.run_callbacks(callbacks, resource_name, *args)
        callbacks = getattr(self, f"{event}_{resource_name}")
        await self.run_callbacks(callbacks, *args)



from .collections import process_collections_request, collections_endpoint_factory
from .documents import process_item_request, item_endpoint_factory
from .endpoints import (
    process_subresource_request,
    subresource_endpoint_factory,
    home_endpoint,
    me_endpoint,
)



from starlette.requests import Request
from fasteve.core.utils import ObjectID
from typing import Union
from fasteve.methods.common import get_document
from fastapi import HTTPException
from fastapi import Response


async def delete(request: Request) -> Response:
    try:
        await request.app.data.remove(request.state.resource)
    except Exception as e:
        raise e
    return Response(status_code=204)


async def delete_item(request: Request, item_id: Union[ObjectID, str]) -> Response:
    document = await get_document(request, item_id)
    if not document:
        raise HTTPException(404)
    try:
        await request.app.data.remove_item(request.state.resource, document["_id"])
    except Exception as e:
        raise e
    return Response(status_code=204)



from starlette.requests import Request
from datetime import datetime
from fasteve.methods.common import get_document
from typing import Union
from fasteve.core.utils import ObjectID
from fastapi import Response


async def patch_item(request: Request, item_id: Union[ObjectID, str]) -> Response:
    orginal_document = await get_document(request, item_id)
    payload = getattr(request, "payload")
    now = datetime.now()
    payload["_created"] = orginal_document["_created"]
    payload["_updated"] = now

    try:
        document = await request.app.data.update_item(
            request.state.resource, orginal_document["_id"], payload
        )
    except Exception as e:
        raise e
    return Response(status_code=204)



from fasteve.io.base import Client, ConnectionException, DataLayer
from fasteve.core import config
from fastapi import HTTPException
from fasteve.resource import Resource
from pymongo.collection import Collection
from motor.motor_asyncio import AsyncIOMotorClient
from fasteve.core.utils import log, ObjectID
from typing import List, Tuple


class DataBase:
    client: AsyncIOMotorClient = None


db = DataBase()


class MongoClient(Client):
    @classmethod
    async def get_database(cls) -> AsyncIOMotorClient:
        return db.client

    @classmethod
    def connect(cls) -> None:
        try:
            client = AsyncIOMotorClient(
                str(config.MONGODB_URI),
                serverSelectionTimeoutMS=config.CONNECTION_TIMEOUT,
            )
        except Exception:
            raise ConnectionException
        db.client = client

    @classmethod
    def close(cls) -> None:
        db.client.close()


class Mongo(DataLayer):
    """MongoDB data access layer for Fasteve."""

    def init_app(self) -> None:
        self.mongo_prefix = None

    async def get_collection(self, resource: Resource) -> Collection:
        # maybe it would be better to use inject db with
        # Depends(get_database) at the path operation function?
        # By better I mean more FastAPI-ish.
        # However, then I have to pass the db all the way down to the
        # datalayer...
        try:
            client = await MongoClient.get_database()
        except Exception as e:
            HTTPException(500, e)
        return client[config.MONGODB_DATABASE][resource.name]

    async def aggregate(
        self,
        resource: Resource,
        pipline: List[dict] = [],
        skip: int = 0,
        limit: int = 0,
    ) -> Tuple[List[dict], int]:
        collection = await self.get_collection(resource)

        paginated_results: List[dict] = []
        paginated_results.append({"$skip": skip})
        paginated_results.append({"$limit": limit})
        facet_pipelines: dict = {}
        facet_pipelines["paginated_results"] = paginated_results
        facet_pipelines["total_count"] = list({"$count": "count"})
        facet = {"$facet": facet_pipelines}
        pipline.append(facet)
        count = 0
        try:
            async for res in collection.aggregate(pipline):
                items = res["paginated_results"]
                if res["total_count"]:
                    # IndexError: list index out of range
                    count = res["total_count"][0]["count"]
        except Exception as e:
            raise e
        return items, count

    async def find(
        self, resource: Resource, query: dict = {}, skip: int = 0, limit: int = 0
    ) -> Tuple[List[dict], int]:
        """Retrieves a set of documents matching a given request. Queries can
        be expressed in two different formats: the mongo query syntax, and the
        python syntax. The first kind of query would look like: ::
            ?where={"name": "john doe"}
        while the second would look like: ::
            ?where=name=="john doe"
        :param resource: Resource object.
        """
        # process_query(q)
        collection = await self.get_collection(resource)
        items = []
        # Perform find and iterate results
        # https://motor.readthedocs.io/en/stable/tutorial-asyncio.html#async-for
        try:
            async for row in collection.find(query, skip=skip, limit=limit):
                items.append(row)
        except Exception as e:
            raise e
        count = await collection.count_documents(query)
        return items, count

    async def find_one(self, resource: Resource, query: dict) -> dict:
        """"""
        collection = await self.get_collection(resource)
        try:
            item = await collection.find_one(query)
        except Exception as e:
            raise e
        return item

    @log
    async def insert(self, resource: Resource, payload: dict) -> dict:
        """"""
        collection = await self.get_collection(resource)
        try:
            await collection.insert_one(payload)
        except Exception as e:
            raise e
        return payload

    async def insert_many(self, resource: Resource, payload: List[dict]) -> List[dict]:
        """"""
        collection = await self.get_collection(resource)
        try:
            await collection.insert_many(payload)
        except Exception as e:
            raise e
        return payload

    async def remove(self, resource: Resource) -> None:
        """Removes an entire set of documents from a
        database collection.
        """
        collection = await self.get_collection(resource)
        try:
            await collection.delete_many({})
        except Exception as e:
            raise e

    async def remove_item(self, resource: Resource, item_id: ObjectID) -> None:
        """Removes a single document from a database collection."""
        collection = await self.get_collection(resource)
        try:
            await collection.delete_one({"_id": item_id})
        except Exception as e:
            raise e

    async def replace_item(
        self, resource: Resource, item_id: ObjectID, payload: dict
    ) -> None:
        """Replaces single document from a database collection"""
        collection = await self.get_collection(resource)
        try:
            await collection.replace_one({"_id": item_id}, payload)
        except Exception as e:
            raise e

    async def update_item(
        self, resource: Resource, item_id: ObjectID, payload: dict
    ) -> None:
        """Updates single document from a database collection"""
        collection = await self.get_collection(resource)
        try:
            await collection.update_one({"_id": item_id}, {"$set": payload})
        except Exception as e:
            raise e



from starlette.requests import Request
from fasteve.core.utils import ObjectID
from typing import Union
from fastapi import HTTPException


async def get_document(request: Request, item_id: Union[ObjectID, str]) -> dict:
    try:
        query = {"_id": ObjectID.validate(item_id)}
    except ValueError:
        # item_id is not a valid ObjectID check if there is an alt_id set
        if not request.state.resource.alt_id:
            raise HTTPException(
                404, "The ObjectID is invaild and there is no alternative ID"
            )
        query = {request.state.resource.alt_id: item_id}
    try:
        document = await request.app.data.find_one(request.state.resource, query)
    except Exception as e:
        raise e
    return document



def render_pymongo_error(details: dict) -> dict:
    key = list(details["keyValue"].keys())[0]
    val = details["keyValue"][key]
    msg = {
        "loc": ["body", "schema", key],
        "msg": f"value '{val}' is not unique",
        "type": "value_error.not_unique",
    }
    return msg



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="fasteve",
    version="0.1.0",
    description="A simple and feature complete REST API framework designed for speed.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Wytamma/fasteve",
    author="Wytamma Wirth",
    author_email="wytamma.wirth@me.com",
    license="BSD",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(exclude=("tests",)),
    install_requires=["fastapi", "uvicorn", "motor", "email-validator"],
)


from starlette.requests import Request
from fasteve.methods import delete, get, post
from fastapi import HTTPException
from fasteve.core import config
from fastapi import Path
from typing import Callable, Optional, Union
from fasteve.resource import Resource, SubResource
from fasteve.core.utils import log, ObjectID
from fasteve.endpoints.documents import process_item_request
from pymongo.errors import DuplicateKeyError, BulkWriteError
from fasteve.io.mongo.utils import render_pymongo_error


@log
async def process_subresource_request(
    request: Request, item_id: Union[ObjectID, str]
) -> dict:
    methods: dict = {"GET": get, "POST": post, "DELETE": delete}
    if request.method not in methods:
        raise HTTPException(405)
    try:
        res = await methods[request.method](request)
        return res
    except DuplicateKeyError as e:
        msg = render_pymongo_error(e.details)
        raise HTTPException(422, msg)
    except BulkWriteError as e:
        msg = render_pymongo_error(e.details["writeErrors"][0])
        raise HTTPException(422, msg)
    except Exception as e:
        raise e


def subresource_endpoint_factory(
    resource: Resource, method: str, subresource: SubResource
) -> Callable:
    # i may have to pass lookup all the way down
    # i think that's okay as I'll have to do it anyway for aggergation

    if method in ("GET", "HEAD"):  # not sure why this...
        # no schema validation on GET request
        if resource.alt_id:

            async def subresource_endpoint_with_alt_id(
                request: Request,
                item_id: Union[ObjectID, str] = Path(
                    ..., alias=f"{resource.item_name}_id"
                ),
            ) -> dict:
                return await process_subresource_request(request, item_id)

            return subresource_endpoint_with_alt_id
        else:

            async def subresource_endpoint(
                request: Request,
                item_id: ObjectID = Path(..., alias=f"{resource.item_name}_id"),
            ) -> Optional[dict]:
                return await process_item_request(request, item_id)

            return subresource_endpoint

    else:
        raise Exception(f'"{method}" is an invalid HTTP method')


def home_endpoint(request: Request) -> dict:
    response = {}
    if config.HATEOAS:  # move to repare response function
        links = []
        for resource in request.app.resources:
            links.append({"href": f"/{resource.name}", "title": f"{resource.name}"})
        response[config.LINKS] = {"child": links}
    return response


def me_endpoint(request: Request) -> dict:
    """for auth / user info"""
    response = {}
    if config.HATEOAS:  # move to prepare response function
        links = []
        for resource in request.app.resources:
            links.append({"href": f"/{resource.name}", "title": f"{resource.name}"})
        response[config.LINKS] = {"child": links}
    return response
